from genanki import Model


BASIC_MODEL = Model(
    1456156020,
    "Basic (Ankivalenz)",
    fields=[
        {
            "name": "Front",
            "font": "Arial",
        },
        {
            "name": "Back",
            "font": "Arial",
        },
        {
            "name": "Path",
            "font": "Arial",
        },
    ],
    templates=[
        {
            "name": "Card 1",
            "qfmt": "<small>{{Path}}</small><br><br>{{Front}}",
            "afmt": "{{FrontSide}}\n\n<hr id=answer>\n\n{{Back}}",
        },
    ],
    css=".card {\n font-family: arial;\n font-size: 20px;\n text-align: center;\n color: black;\n background-color: white;\n}\n",
)

BASIC_AND_REVERSED_CARD_MODEL = Model(
    2105966946,
    "Basic (and reversed card) (Ankivalenz)",
    fields=[
        {
            "name": "Front",
            "font": "Arial",
        },
        {
            "name": "Back",
            "font": "Arial",
        },
        {
            "name": "Path",
            "font": "Arial",
        },
    ],
    templates=[
        {
            "name": "Card 1",
            "qfmt": "<small>{{Path}}</small><br><br>{{Front}}",
            "afmt": "{{FrontSide}}\n\n<hr id=answer>\n\n{{Back}}",
        },
        {
            "name": "Card 2",
            "qfmt": "<small>{{Path}}</small><br><br>{{Back}}",
            "afmt": "{{FrontSide}}\n\n<hr id=answer>\n\n{{Front}}",
        },
    ],
    css=".card {\n font-family: arial;\n font-size: 20px;\n text-align: center;\n color: black;\n background-color: white;\n}\n",
)

CLOZE_MODEL = Model(
    1642839314,
    "Cloze (Ankivalenz)",
    model_type=Model.CLOZE,
    fields=[
        {
            "name": "Text",
            "font": "Arial",
        },
        {
            "name": "Back Extra",
            "font": "Arial",
        },
        {
            "name": "Path",
            "font": "Arial",
        },
    ],
    templates=[
        {
            "name": "Cloze",
            "qfmt": "<small>{{Path}}</small><br><br>{{cloze:Text}}",
            "afmt": "<small>{{Path}}</small><br><br>{{cloze:Text}}<br>\n{{Back Extra}}",
        },
    ],
    css=".card {\n font-family: arial;\n font-size: 20px;\n text-align: center;\n color: black;\n background-color: white;\n}\n\n"
    ".cloze {\n font-weight: bold;\n color: blue;\n}\n.nightMode .cloze {\n color: lightblue;\n}",
)



from typing import List
from .types import Node, Delimeter, BasicCard, Card, ClozeCard


class NodeParser:
    def parse(
        self, nodes: Node, cards: List[Card] = None, path: List[str] = None
    ) -> List[Card]:
        if cards is None:
            cards = []

        if path is None:
            path = []

        match nodes:
            case list(l):
                for node in l:
                    self.parse(node, cards, path.copy())
            case (str(header), list(l)):
                path.append(header)
                return self.parse(l, cards, path.copy())
            case str(question):
                return cards.append(ClozeCard(question, path.copy()))
            case (Delimeter("::"), str(answer)):
                return cards.append(
                    BasicCard(path[-1], answer, path[:-1], reverse=True)
                )
            case (Delimeter("?::"), str(answer)):
                return cards.append(BasicCard(path[-1], answer, path[:-1]))
            case (Delimeter("::?"), str(question)):
                return cards.append(BasicCard(question, path[-1], path[:-1]))
            case (str(question), Delimeter("?::"), str(answer)):
                return cards.extend([BasicCard(question, answer, path.copy())])
            case (str(answer), Delimeter("::?"), str(question)):
                return cards.extend([BasicCard(question, answer, path.copy())])
            case (str(question), Delimeter("::"), str(answer)):
                return cards.extend(
                    [BasicCard(question, answer, path.copy(), reverse=True)]
                )

        return cards



from dataclasses import dataclass, field
from typing import Tuple, Union


@dataclass
class Delimeter:
    contents: str


PathNode = Tuple[str, list["Node"]]
BasicNode = Tuple[str, Delimeter, str]
StandaloneNode = Tuple[Delimeter, str]
ClozeNode = str

Node = Union[list["Node"], PathNode, BasicNode, StandaloneNode, ClozeNode]


@dataclass(order=True)
class BasicCard:
    question: str
    answer: str
    path: list[str]
    reverse: bool = False


@dataclass(order=True)
class ClozeCard:
    question: str
    path: list[str]


Card = Union[BasicCard, ClozeCard]

Path = list[str]



from datetime import datetime
import json
import os
import pathlib
import random
from typing import List, Optional

from ankivalenz import HtmlParser, NodeParser
from .anki_models import BASIC_AND_REVERSED_CARD_MODEL, BASIC_MODEL, CLOZE_MODEL
from .types import BasicCard, Card, ClozeCard, Path
import genanki


class Note(genanki.Note):
    pass


def format_path(path: Path) -> str:
    return " > ".join(path)


def cards_to_notes(cards: List[Card], time: Optional[datetime]) -> List[Note]:
    if time is None:
        time = datetime.now()

    notes = []

    updated_tag = "ankivalenz:updated:{}".format(int(time.timestamp()))

    for card in cards:
        if isinstance(card, BasicCard):
            note = Note(
                model=BASIC_AND_REVERSED_CARD_MODEL if card.reverse else BASIC_MODEL,
                fields=[card.question, card.answer, format_path(card.path)],
                # Add tag `ankivalenz:updated:<time in epoch>` to note.
                tags=[updated_tag],
            )
            notes.append(note)
        if isinstance(card, ClozeCard):
            note = Note(
                model=CLOZE_MODEL,
                fields=[card.question, "", format_path(card.path)],
                tags=[updated_tag],
            )
            notes.append(note)

    return notes


def load_cards(path: pathlib.Path) -> List[Card]:
    cards = []

    for file in path.glob("**/*.html"):
        with file.open() as f:
            nodes = HtmlParser().parse(f.read())
            cards.extend(NodeParser().parse(nodes))

    return cards


def init(path: pathlib.Path) -> str:
    deck_id = random.randrange(1 << 30, 1 << 31)
    # set deck_name to the name of the directory:
    deck_name = path.name
    json_path = path / "ankivalenz.json"

    with open(json_path, "w") as f:
        json.dump(
            {"deck_id": deck_id, "deck_name": deck_name},
            f,
            indent=4,
            sort_keys=True,
        )

    return json_path


def package(path: pathlib.Path, time: Optional[datetime] = None) -> genanki.Package:
    with open(os.path.join(path, "ankivalenz.json")) as f:
        settings = json.load(f)

    cards = load_cards(path)

    deck = genanki.Deck(
        settings["deck_id"],
        settings["deck_name"],
    )

    for note in cards_to_notes(cards, time=time):
        deck.add_note(note)

    package = genanki.Package(deck)

    for file in path.glob("images/*"):
        package.media_files.append(str(file))

    return package



from datetime import datetime
import pathlib
import typer
from . import generator

app = typer.Typer()


@app.command()
def run(path: str):
    time = datetime.now()
    full_path = pathlib.Path(path)
    package = generator.package(full_path, time=time)

    apkg_path = package.decks[0].name + ".apkg"
    package.write_to_file(apkg_path)

    typer.echo(
        "- Added {} notes to deck {} in {}".format(
            len(package.decks[0].notes), package.decks[0].name, apkg_path
        )
    )
    typer.echo("- Import the .apkg file into Anki (File -> Import)")
    typer.echo("- Find and delete orphaned notes with this filter (Browse):")
    typer.echo(
        "    deck:{} -tag:ankivalenz:updated:{}".format(
            package.decks[0].name, int(time.timestamp())
        )
    )


@app.command()
def init(path: str):
    full_path = pathlib.Path(path).resolve()
    json_path = generator.init(full_path)

    typer.echo("Created ankivalenz.json at {}".format(json_path))



import itertools
import re
from typing import List
from bs4 import BeautifulSoup, NavigableString, PageElement, Tag

from .types import ClozeNode, Delimeter, Node

BASIC_REGEXP = r"(.*?)\s*(::\?|\?::|::)\s*(.*)"
STANDALONE_REGEXP = r"^(::\?|\?::|::)\s*(.*)"
CLOZE_REGEXP = r".*{{c\d+::.*?}}.*"
HEADER_TAGS = ["h1", "h2", "h3", "h4", "h5", "h6"]

# Returns `True` if `element` is a non-header or a header with
# a lower priority than `header`, i.e. a "descendant", hierarchically
# speaking
def is_descendant(header: str, element: PageElement) -> bool:
    if not (isinstance(element, Tag)):
        return True

    if not (element.name in HEADER_TAGS):
        return True

    return element.name > header


class HtmlParser:
    def parse(self, text: str) -> List[Node]:
        soup = BeautifulSoup(text, "html.parser")

        self.strip_image_paths(soup)

        return self.find_nodes(soup.contents)

    # recursively traverse the tree, and modify all image srcs to be the basename only:
    def strip_image_paths(self, elements: List[PageElement]) -> None:
        for i, e in enumerate(elements):
            if isinstance(e, Tag):
                if e.name == "img":
                    e["src"] = e["src"].split("/")[-1]
                self.strip_image_paths(e.contents)

    def find_nodes(self, elements: List[PageElement]) -> List[Node]:
        nodes: List[Node] = []

        for idx, element in enumerate(elements):
            if isinstance(element, Tag):
                if element.name in HEADER_TAGS:
                    header = "".join(map(str, element.contents))

                    children = [
                        *itertools.takewhile(
                            lambda x: is_descendant(element.name, x),
                            elements[idx + 1 :],
                        )
                    ]
                    rest = [
                        *itertools.dropwhile(
                            lambda x: is_descendant(element.name, x),
                            elements[idx + 1 :],
                        )
                    ]

                    nodes.append((header, self.find_nodes(children)))
                    nodes.extend(self.find_nodes(rest))
                    break
                elif element.name == "li":
                    head = [
                        *itertools.takewhile(
                            lambda x: not (
                                isinstance(x, Tag) and x.name in ["ul", "ol"]
                            ),
                            element.children,
                        )
                    ]
                    tail = [
                        *itertools.dropwhile(
                            lambda x: not (
                                isinstance(x, Tag) and x.name in ["ul", "ol"]
                            ),
                            element.children,
                        )
                    ]

                    # if there are any elements in `tail`, it means there is a nested list.
                    if len(tail) > 0:
                        # in the case of a nested list, the header is the content of
                        # the first tag in `head`. If we do not find a tag in `head`,
                        # we do not consider the nested list, and simply skip it.
                        for e in head:
                            if isinstance(e, Tag):
                                if e.name == "p":
                                    header = "".join(map(str, e.contents))
                                else:
                                    header = str(e)

                                nodes.append((header, self.find_nodes(tail)))
                                break
                    else:
                        nodes.extend(self.find_nodes(head))
                else:
                    nodes.extend(self.find_nodes(element.contents))
            elif isinstance(element, NavigableString):
                if match := re.match(CLOZE_REGEXP, str(element), re.DOTALL):
                    nodes.append(str(element).strip())
                    break
                elif match := re.match(STANDALONE_REGEXP, str(element), re.DOTALL):
                    delimeter_type, after_match = match.groups()

                    after = after_match + "".join(map(str, elements[idx + 1 :]))

                    nodes.append((Delimeter(delimeter_type), after.strip()))
                    break
                elif match := re.match(BASIC_REGEXP, str(element), re.DOTALL):
                    before_match, delimeter_type, after_match = match.groups()

                    before = "".join(map(str, elements[:idx])) + before_match
                    after = after_match + "".join(map(str, elements[idx + 1 :]))

                    nodes.append(
                        (before.strip(), Delimeter(delimeter_type), after.strip())
                    )
                    break

        return nodes



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ankivalenz']

package_data = \
{'': ['*']}

install_requires = \
['bs4>=0.0.1,<0.0.2', 'genanki>=0.13.0,<0.14.0', 'typer>=0.6.1,<0.7.0']

entry_points = \
{'console_scripts': ['ankivalenz = ankivalenz.main:app']}

setup_kwargs = {
    'name': 'ankivalenz',
    'version': '0.1.0',
    'description': '',
    'long_description': "# Ankivalenz\n\nAnkivalenz is a tool for generating Anki notes from HTML files.\n\n## Tutorial\n\nIn this walk-through we will write our notes as Markdown files, use\npandoc[^pandoc] to convert them to HTML, and finally use Ankivalenz to\ngenerate an Anki deck with Anki notes extracted from our Markdown files.\n\n### Installation\n\nAnkivalenz is distributed as a Python package, and requires Python 3.10+. To install:\n\n```\n$ pip3 install ankivalenz\n```\n\n### Initialize project\n\nCreate a folder for your notes:\n\n```\n$ mkdir Notes\n$ cd Notes\n```\n\nAnkivalenz needs a configuration file, containing the name and ID of the\nAnki deck. This can be generated with `ankivalenz init`:\n\n```\n$ ankivalenz init\n```\n\n### Write a note\n\nAdd the following to a file named `Cell.md`:\n\n```markdown\n# Cell\n\n## Types\n\n- Prokaryotic ?:: does not contain a nucleus\n- Eukaryotic ?:: contains a nucleus\n```\n\n### Generate Anki deck\n\nConvert it to HTML:\n\n```\n$ pandoc Cell.md > Cell.html\n```\n\nAnd run Ankivalenz:\n\n```\n$ ankivalenz run .\n```\n\nThis generates a file `Notes.apkg` that can be imported to Anki. Open\nAnki and go to File -> Import, and find `Notes.apkg`.\n\n### Review\n\nThe new Anki deck will have two notes:\n\n| Question    | Answer                     | Path         |\n| ----------- | -------------------------- | ------------ |\n| Prokaryotic | does not contain a nucleus | Cell > Types |\n| Eukaryotic  | contains a nucleus         | Cell > Types |\n\nThis is what the first note looks like in Anki:\n\n![Anki review](images/anki-review.png)\n\n## Syntax\n\n### Front/back cards\n\nAnkivalenz supports front/back cards, where the front is the question\nand the back is the answer. To create a front/back card, add a new list item\nwith the question, followed by `?::` and the answer:\n\n```markdown\n- Color of the sun ?:: Yellow\n```\n\nYou can flip the order of the question and answer by using `::?` instead:\n\n```markdown\n- Anwer ::? Question\n```\n\n#### Two-way notes\n\nTwo-way notes can be created with `::`:\n\n```markdown\n- Side 1 :: Side 2\n```\n\nThis will create two cards in Anki:\n\n| Front  | Back   |\n| ------ | ------ |\n| Side 1 | Side 2 |\n| Side 2 | Side 1 |\n\n### Cloze cards\n\nAnkivalenz supports cloze deletion[^cloze], where the answer is hidden in the\nquestion. To create a cloze card, add a new list item with the question,\nusing Anki's cloze syntax:\n\n```markdown\n- The {{c1::sun}} is {{c2::yellow}}.\n```\n\n### Nested lists\n\nLists can be nested:\n\n```markdown\n- Solar System\n  - Star ?:: Sun\n  - Planet\n    - Earth ?:: Blue\n    - Mars ?:: Red\n```\n\nThe headings for the nested lists become a part of the notes' paths:\n\n| Question | Answer | Path                  |\n| -------- | ------ | --------------------- |\n| Star     | Sun    | Solar System          |\n| Earth    | Blue   | Solar System > Planet |\n| Mars     | Red    | Solar System > Planet |\n\n### Math\n\nIf you are writing Markdown files, and use pandoc to convert them,\nthe following syntax for math is supported:\n\n```markdown\n- Inline math: $1 + 2$\n- Display math: $$1 + 2$$\n```\n\nWith the `--mathjax` flag, pandoc will generate the correct markup,\nusing `\\( ... \\)` as delimeters for inline math, and `\\[ ... \\]` as\ndelimeters for display math:\n\n```\n$ pandoc --mathjax Note.md > Note.html\n```\n\n[^pandoc]: https://pandoc.org/\n[^cloze]: https://docs.ankiweb.net/editing.html#cloze-deletion\n",
    'author': 'Harry Vangberg',
    'author_email': 'harry@vangberg.name',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
